// Adds an inotify channel to the main loop, using giochannels.
// There is just one inotify instance (which can watch multiple paths)
// connected to a g_io_channel.
//
// Code based on the following:
// inotify docs:    http://www.linuxjournal.com/article/8478 (Robert Love)
// iochannels docs: http://www.linuxjournal.com/article/8545 (Robert Love)
// May 05 example code:
// http://www.kernel.org/pub/linux/kernel/people/rml/inotify/glib/inotify-glib-0.0.3.tar.gz
// buffer for slurping inotify events

#include <gtk/gtk.h>
#include <lauxlib.h>
#include "local_inotify.h"
#include "local_inotify_syscalls.h"
#include "utils.h"
#include "notify.h"

#define INOTIFY_BUFSZ  32768

static int inotify_fd = 0;  // just one fd for all inotify watches

/// cb for giochannel that is watching the inotify fd
static gboolean watch_handler (GIOChannel *gio, GIOCondition UNUSED(condition), gpointer data) {
	char buf[INOTIFY_BUFSZ];
	guint len;
	guint i = 0;
	lua_State *L = data;
	dbg("entering watch handler\n");
	// read in as many pending events as we can
	
#if 0
	 // I had this working before I found the NONBLOCK flag needed with READ_CHARS
	GIOError err = g_io_channel_read (gio, buf, INOTIFY_BUFSZ, &len);
	if (err != G_IO_ERROR_NONE) {
		warn("Error reading inotify fd: %d\n", err);
		return 1;
	}
#else
	GIOStatus ret = g_io_channel_read_chars(gio, buf, INOTIFY_BUFSZ, &len, NULL);
	if (ret != G_IO_STATUS_NORMAL) {
		warn("result reading chars inotify fd: %d\n", ret);
		return 1;
	}
#endif
	
	// reconstruct each event and send to the user's callback
	gboolean again;
	while (i < len) {
		const char *name = "noname";
		struct inotify_event *event;
		event = (struct inotify_event *) &buf[i];
		if (event->len) {
			name = &buf[i] + sizeof (struct inotify_event);
		}
		int tos = lua_gettop(L);  // preserve stack level // {as is}
		if (!lua_checkstack(L, 7)) {
			warn("no stack for notify cb\n");
			return 1;
		}
		lua_rawgeti(L, LUA_ENVIRONINDEX, event->wd);      // t
		lua_getfield(L, -1, "func");                      // t f
		lua_pushstring(L, name);                          // t f n
		lua_pushinteger(L, event->mask);                  // t f n m
		lua_pushinteger(L, event->wd);                    // t f n m w
		lua_pushinteger(L, event->cookie);                // t f n m w c
		lua_getfield(L, -6, "udat");                      // t f n m w c u
		lua_call(L, 5 ,1);                                // again
		again = lua_toboolean(L, -1);                     // again
		lua_settop(L, tos);                               // {as was}
		i += sizeof (struct inotify_event) + event->len;
	}
	return again;
}

int l_addwatch(lua_State *L) {  // path, mask, luafunc (udata?)
	const char* watchpath = luaL_checkstring(L, 1);
	int watchmask = luaL_checkinteger(L, 2);
	luaL_checktype(L, 3, LUA_TFUNCTION);
	lua_settop(L, 4);                         // path mask luafunc udata/nil
	
	// If necessary, make our (only) inotify instance and connect it to a gio channel
	if (inotify_fd == 0 ) {  
		inotify_fd = inotify_init();
		if (inotify_fd <= 0) {
			warn("inotify_init failed");
			return 0;
		}
		GIOChannel *gio = g_io_channel_unix_new(inotify_fd);
		if (!gio) {
			warn("giochannel creation failed");
			return 0;
		}
		g_io_channel_set_encoding(gio, NULL, NULL);
		// setting NONBLOCK is critical if we use g_io_channel_read_chars
		g_io_channel_set_flags(gio, G_IO_FLAG_NONBLOCK, NULL);
		if (!g_io_add_watch (gio, G_IO_IN, watch_handler, L)) {
			warn("connecting inotify fd to giochannel failed");
			return 0;
		}
	}
	
	int wd = inotify_add_watch(inotify_fd, watchpath, watchmask);
	if (wd < 0) {
		warn("inotify_add_watch failed");
		return 0;
	}
	
	// save data for callback time in the environment, indexed by wd.
	lua_createtable(L, 2, 0);       // path mask luafunc udata t
	lua_pushvalue(L, 3);            // path mask luafunc udata t luafunc
	lua_setfield(L, -2, "func");    // path mask luafunc udata t  {t[func]=luafunc}
	lua_pushvalue(L, 4);            // path mask luafunc udata t udata
	lua_setfield(L, -2, "udat");    // path mask luafunc udata t  {t[func]=luafunc}
	lua_pushinteger(L, wd);         // path mask luafunc udata t wd
	lua_insert(L, -2);              // path mask luafunc udata wd t
	lua_settable(L, LUA_ENVIRONINDEX); // path mask luafunc udata {E[wd]=t}
	
	dbg("watch %d set for path %s\n", wd, watchpath);
	lua_pushinteger(L, wd);
	return 1;
}

int l_removewatch(lua_State *L) {
	int wd = luaL_checkinteger(L, 1);
	int ret = inotify_rm_watch(inotify_fd, wd);
	if (ret < 0) {
		warn("inotify_rm_watch failed\n");
	}
	return 0;
}

void set_notify_masks(lua_State *L) {
	defconst(IN_ACCESS);
	defconst(IN_MODIFY);
	defconst(IN_ATTRIB);
	defconst(IN_CLOSE_WRITE);
	defconst(IN_CLOSE_NOWRITE);
	defconst(IN_OPEN);
	defconst(IN_MOVED_FROM);
	defconst(IN_MOVED_TO);
	defconst(IN_CREATE);
	defconst(IN_DELETE);
	defconst(IN_DELETE_SELF);
	defconst(IN_UNMOUNT);
	defconst(IN_Q_OVERFLOW);
	defconst(IN_IGNORED);
	defconst(IN_CLOSE);
	defconst(IN_MOVE);
	defconst(IN_ISDIR);
	defconst(IN_ONESHOT);
	defconst(IN_ALL_EVENTS);
}
