#!/usr/bin/lua
-- Find certain functions in a file of declarations from stdin,
-- compute signatures for them using the list of argument type -> nibble code
-- conversions in typecodes{}, and print a list of funcname=sig lines for the
-- arbitrary hash function.

-- The input file must be in the format produced by the gcc -aux-info option.

-- This version supports four 4-bit fields for the signature (which thus fits into
-- a short int)  <a3><a2><a1><ret> is the bitmap, where ret is the return type
-- and a1,a2,a4 are up to three arguments. The 4-bit encoding is specified in
-- typecodes{}.  ai == 0 means there is no arg in that position.

-- bring in typecodes = {void=0, bool=1, int=2, ...  and ERR* constants
require "typetables"    

-- Parse the line into func name, ret type, arg list
funcs = {}
for line in io.lines() do
	-- ignore everything except extern declarations
	decl = line:match("^/%* /usr/include/gtk%-2%.0/gtk/.* extern (.+);")
	if decl then
		ret, funcname, args = decl:match("^(.+)%s(%S+)%s%((.+)%)")
		arglist={}
		for arg in args:gmatch("[^,]+") do            -- Split arg list on commas
			arglist[#arglist+1] = arg:match(" ?(.*)")  -- Drop leading space
		end
		
		-- Move stars from start of func name to end of return type
		startstars, endstars, removedstars= funcname:find("^(%*+)")
		if startstars then
			funcname = funcname:sub(endstars+1)
			ret =  ret.." "..removedstars
		end
		
		funcs[#funcs+1] = {funcname, ret, arglist}
	end
end

-- Select those functions that we want to put into the hash.  We want:
-- (1) Methods, defined as start of function name matching first argument e.g.
-- GtkSpinButton * gtk_spin_button_set_range   is a method of class GtkSpinButton.
-- (2) Certain other functions like gtk_main().
-- In the case of methods, this step modifies the argument list: it deletes the
-- first argument, and inserts void if there is no other argument.  This is because
-- there is no need record the first argument of a method - it must be a widget*.
-- Thus we keep room for three useful arguments (plus the return type) in the
-- 4-nibble hash.
otherfuncs = {gtk_main=true, gtk_main_quit=true}
wanted = {}
for _,func in ipairs(funcs) do
	funcname = func[1]
	rettype = func[2]
	args = func[3]
	classname = args[1]:match("^(.+) %*")
	if classname then
		-- change ClassName to class_name and treat any function that begins
		-- with this as a method.
		funcprefix = classname:gsub("(%l)(%u)", "%1_%2"):lower()
		--CHANGED if funcname:find(funcprefix) then   to...
		if funcname:find("^"..funcprefix.."_") then
			table.remove(args,1)
			if #args == 0 then
				args[1] = "void"
			end
			wanted[#wanted+1] = {funcname, rettype, args}
		end
	end
	
	-- Get ctors in general, unmodified
	if funcname:find("^gtk_%a+_new$") then
		wanted[#wanted+1] = {funcname, rettype, args}
	end
	
	-- Get other wanted functions and add those unmodified.
	if otherfuncs[funcname]  then
		wanted[#wanted+1] = {funcname, rettype, args}
	end
end

-- Sort the wanted function list by name
table.sort(wanted, function (a,b) return (a[1] < b[1]) end)

-- Make a short int representing the types of input and output parameters.
function hashvalue(rettype, paramlist)
	if #paramlist > 3 then return ERRNARGS end
	
	totalval = 0
	val = typecodes[rettype]
	if not val then return ERRTYPE end
	totalval = totalval + val
	
	for i, p in pairs(paramlist) do
		val = typecodes[p]
		if not val then return ERRTYPE end
		totalval = totalval + val * 2^(4*i)
	end
	
	return totalval
end

-- Print wanted function signatures in a form suitable for the mph arbitrary
-- hash generator: funcname=9876
for _,func in ipairs(wanted) do
	io.write(string.format("%s=%d\n", func[1], hashvalue(func[2], func[3])))
end



--[[
-- Print the parsed method table
for _,func in ipairs(wanted) do
	io.write(string.format("%s|%s|", func[1], func[2]))
	for _,argtype in ipairs(func[3]) do
		io.write(string.format("%s|", argtype))
	end
	io.write("\n")
end
]]

